@@ -26,6 +26,7 @@ class LoginCookie:
     The LoginCookie object can be used in place of a password for the :meth:`client.CloudClient.run` method.
 
     Cookie Example Usage::
+
         from scratchcloud import CloudClient, LoginCookie
 
         login_cookie = LoginCookie(
@@ -36,6 +37,7 @@ class LoginCookie:
         client = CloudClient('username', '123')
         # more code here
         client.run(login_cookie)
+
     """
 
     def __init__(self, csrftoken: str, sessionid: str):
@@ -129,6 +131,9 @@ class CloudClient:
     :param event_loop: the event loop that the CloudClient will use.
         defaults to asyncio.get_event_loop()
     :type event_loop: AbstractEventLoop
+    :param ignore_missing_variables: prevents a MissingCloudVariable exception during the websocket handshake if cloud variables are missing from the connected project.
+        defaults to False
+    :type: bool
 
     **Attributes:**
     
@@ -158,7 +163,7 @@ class CloudClient:
     **Methods:**
     """
 
-    def __init__(self, username: str, project_id: str, max_reconnect: int = None, reconnect_cooldown: int = 10, encoder: Callable[[str], str] = None, decoder: Callable[[str], str] = None, disconnect_messages: bool = False, max_cache_length: int = 1000, event_loop = asyncio.get_event_loop()):
+    def __init__(self, username: str, project_id: str, max_reconnect: int = None, reconnect_cooldown: int = 10, encoder: Callable[[str], str] = None, decoder: Callable[[str], str] = None, disconnect_messages: bool = False, max_cache_length: int = 1000, event_loop = asyncio.get_event_loop(), ignore_missing_variables: bool = False):
 
         self.username = username
         self.project_id = project_id
@@ -170,12 +175,14 @@ class CloudClient:
         self.headers = None
         self.logged_in = False
         self.ws = None
-        self.connected = False
+        self.client_setup = False
 
         self.max_reconnect = max_reconnect
         self.reconnect_cooldown = reconnect_cooldown
         self.disconnect_messages = disconnect_messages
 
+        self.ignore_missing_variables = ignore_missing_variables
+
         self.decoder = decoder
         self.encoder = encoder
 
@@ -188,6 +195,10 @@ class CloudClient:
 
         self.on_message_registered = False
 
+        self.next_set_time: float = 0
+
+        self.run_client_task = None
+
     # RUNNING CLIENT
     def run(self, token: str | LoginCookie):
         """A blocking function to run the client with library reconnecting.
@@ -222,7 +233,8 @@ class CloudClient:
         while True:
             try:
                 # Run Main loop
-                loop.run_until_complete(self.start(token))
+                self.run_client_task = loop.run_until_complete(self.setup(token))
+                loop.run_until_complete(self.run_client_task)
             except KeyboardInterrupt:
                 # Stop Loop if KeyboardInterrupt
                 loop.create_task(self.on_disconnect_task())
@@ -234,7 +246,7 @@ class CloudClient:
                     print(f'Disconnected due to type: {type(e)}\n{e}')
                 
                 # If previously connected, run disconnect task and reconnect again.
-                if self.connected:
+                if self.client_setup:
                     restart = True
                     reconnects = 0
                 else:
@@ -256,15 +268,20 @@ class CloudClient:
                 if reconnects == self.max_reconnect:
                     print(f'Reconnection failed {reconnects} times. Stopping...')
                     raise e
+            
+            except asyncio.CancelledError:
+                loop.run_until_complete(self.close())
+                break
+                
             except Exception as e:
                 print(f'ScratchCloud got uncaught Exception with type: {e}')
                 raise e
 
-    async def start(self, token: str | LoginCookie):
-        """A coroutine that starts a client.
+    async def setup(self, token: str | LoginCookie) -> asyncio.Task:
+        """A coroutine that sets up a client.
 
-        Calls all functions needed to connect to scratch in chronological order.
-        Closes self, logs in, connects to the websocket, performs a handshake, and finally, runs the client.
+        Calls all functions needed to connect to scratch in chronological order and returns a run client task.
+        Closes self, logs in, connects to the websocket, and performs a handshake.
 
         This can be used in place of :meth:`run` if you want more control over the event loop.
 
@@ -273,19 +290,21 @@ class CloudClient:
         :param token: the cookie for the account that will be used to establish a connection.
         :type token: LoginCookie
 
+        :rtype: :class:`asyncio.Task`
+
         Internal.
         """
 
-        self.connected = False
+        self.client_setup = False
         await self.close()
 
         await self.login(token)
         await self.connect_ws()
         await self.ws_handshake()
         
-        self.connected = True
-        await self.run_client()
-
+        self.client_setup = True
+        
+        return self.loop.create_task(self.run_client())
 
     async def run_client(self):
         """A coroutine that calls the on_connect_task and starts receving data from the websocket.
@@ -296,6 +315,20 @@ class CloudClient:
 
         await asyncio.gather(self.on_connect_task(), self.on_recv())
 
+    def stop(self):
+        """ A function that stops the client by cancelling the `run_client_task` task.
+
+        Example::
+            
+            @client.event
+            async def on_connect():
+                print('I got here! Time to go...')
+                client.stop()
+        """
+
+        if self.run_client_task is not None:
+            self.run_client_task.cancel()
+
     async def close(self):
         """A coroutine that closes self.http_session.
 
@@ -396,6 +429,8 @@ class CloudClient:
         try:
             data = await asyncio.wait_for(self.ws.recv(), 5)
         except:
+            if self.ignore_missing_variables:
+                return
             await self.close()
             raise MissingCloudVariable('No Cloud Variables Found!')
         self.cloud_variables.update(self.parse_raw_cloud(data))
@@ -698,8 +733,6 @@ class CloudClient:
         Note that above `client.set_cloud` is called from the `on_connect()` coroutine
         """
 
-        value = str(value)
-
         if self.encoder and encode:
             value = self.encoder(value)
         
@@ -717,7 +750,15 @@ class CloudClient:
             'project_id': self.project_id,
         }
 
+        # Ratelimit Handler (1 request every 0.1 seconds)
+
+        if not time.time() >= self.next_set_time:
+            await asyncio.sleep(self.next_set_time - time.time())
+
         await self.ws_send(payload)
+        self.next_set_time = time.time() + 0.1
+        
+
         if name in self.cloud_variables:
             prev = self.cloud_variables[name]
         else:
